from django.contrib.auth.models import AbstractUser
from django.db import models

from drf_anonymous_login.models import AnonymousLoginUserMixin


class PublicModel(models.Model):
    """
    Public model that is visible to anyone (no login required)
    """

    name = models.CharField(max_length=50, primary_key=True)


class PrivateModel(models.Model):
    """
    Private model that requires valid login to be shown
    """

    name = models.CharField(max_length=50, primary_key=True)


class User(AnonymousLoginUserMixin, AbstractUser):
    pass



from django.contrib.auth.models import User
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from rest_framework import authentication, exceptions

from .models import AnonymousLogin

AUTH_KEYWORD = "Token"
AUTH_HEADER = "HTTP_X_AUTHORIZATION_ANONYMOUS"
AUTH_COOKIE = "anonymous_token"


class AnonymousLoginAuthentication(authentication.BaseAuthentication):
    keyword = AUTH_KEYWORD

    def authenticate(self, request):
        auth = request.META.get(AUTH_HEADER, "").split() or request.COOKIES.get(AUTH_COOKIE, "").split()

        if not auth or auth[0].lower() != self.keyword.lower():
            return None

        if len(auth) == 1:
            msg = _("Invalid token header. No credentials provided.")
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _("Invalid token header. Token string should not contain spaces.")
            raise exceptions.AuthenticationFailed(msg)

        try:
            anonymous_login = AnonymousLogin.objects.get(token=auth[1])
            if anonymous_login.expiration_datetime and anonymous_login.expiration_datetime <= timezone.now():
                raise exceptions.AuthenticationFailed("Expired authentication token")
        except AnonymousLogin.DoesNotExist as exc:
            raise exceptions.AuthenticationFailed("Invalid authentication token") from exc

        return User(username=anonymous_login.token), None



# Generated by Django 3.2.18 on 2023-04-03 12:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("drf_anonymous_login", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="anonymouslogin",
            name="expiration_datetime",
            field=models.DateTimeField(
                default=None,
                null=True,
                verbose_name="expiration datetime",
            ),
        ),
        migrations.AlterField(
            model_name="anonymouslogin",
            name="created",
            field=models.DateTimeField(auto_now_add=True, verbose_name="created"),
        ),
        migrations.AlterField(
            model_name="anonymouslogin",
            name="token",
            field=models.CharField(
                db_index=True,
                max_length=64,
                unique=True,
                verbose_name="token",
            ),
        ),
    ]



from django.contrib import admin

from .models import AnonymousLogin


@admin.register(AnonymousLogin)
class AnonymousLoginAdmin(admin.ModelAdmin):
    list_display = ("token", "created")
    readonly_fields = ("token", "created", "request_data")



from django.contrib import admin
from django.urls import include, path

from rest_framework import routers

from testapp.views import PrivateModelViewSet, PublicModelViewSet

from drf_anonymous_login.views import CreateAnonymousLoginViewSet

router = routers.DefaultRouter()
# app routes
router.register(r"public_models", PublicModelViewSet)
router.register(r"private_models", PrivateModelViewSet)

# anonymous login route
router.register(
    r"auth_anonymous",
    CreateAnonymousLoginViewSet,
    basename="auth_anonymous",
)

urlpatterns = [path("admin/", admin.site.urls), path("api/", include(router.urls))]



import re

from rest_framework import (
    filters,
    mixins,
    pagination,
    permissions,
    serializers,
    status,
    viewsets,
)
from rest_framework.response import Response

from drf_anonymous_login.authentication import AnonymousLoginAuthentication
from drf_anonymous_login.models import AnonymousLogin


class AnonymousLoginSerializer(serializers.Serializer):
    token = serializers.CharField(read_only=True)

    class Meta:
        fields = ("token",)


class CreateAnonymousLoginViewSet(mixins.CreateModelMixin, viewsets.GenericViewSet):
    serializer_class = AnonymousLoginSerializer
    authentication_classes = ()
    permission_classes = (permissions.AllowAny,)

    @staticmethod
    def extract_request_headers(request):
        regex = re.compile("^HTTP_")
        return {regex.sub("", header): value for (header, value) in request.META.items() if header.startswith("HTTP_")}

    def create(self, request, *args, **kwargs):
        user = AnonymousLogin.objects.create(
            request_data={
                "data": request.data,
                "headers": self.extract_request_headers(request),
            },
        )
        response = Response({"token": user.token}, status=status.HTTP_201_CREATED)
        response.set_cookie("anonymous_token", f"Token {user.token}")
        return response


class AnonymousLoginAuthenticationModelViewSet(viewsets.ModelViewSet):
    filter_backends = (filters.SearchFilter,)
    authentication_classes = (AnonymousLoginAuthentication,)
    permission_classes = (permissions.IsAuthenticated,)
    pagination_class = pagination.LimitOffsetPagination



import signal
import time

from django import db
from django.core.management.base import BaseCommand
from django.utils import timezone

from drf_anonymous_login.models import AnonymousLogin


class Command(BaseCommand):
    help = "Cleanup (remove) any expired AnonymousLogin instances"

    def _handle_options(self, options):
        self.tick = options["tick"]

    def _handle_termination(self, sig, frame):
        self._running = False

    def handle(self, *args, **options):
        # Load given options.
        self._handle_options(options)

        while self._running:
            time.sleep(self.tick)

            try:
                self.handle_tick()

            except Exception as exc:
                self.stdout.write(
                    f"{exc.__class__.__name__} exception occurred ... ",
                    ending="",
                )
                self.stdout.flush()
                self.stdout.write(self.style.ERROR("FATAL"))
                self.stdout.flush()

                # As the database connection might have failed, we discard it here, so django will
                # create a new one on the next database access.
                db.close_old_connections()

    @staticmethod
    def handle_tick():
        # get all AnonymousLogin that already expired and delete them
        AnonymousLogin.objects.filter(
            expiration_datetime__isnull=False,
            expiration_datetime__lte=timezone.now(),
        ).delete()

    def add_arguments(self, parser):
        # Optional argument for setting the sleep time in seconds after each
        # iteration (a.k.a. tick time).
        parser.add_argument(
            "--tick",
            action="store",
            dest="tick",
            type=int,
            default=1,
            help="Sleep time after each iteration in seconds.",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The command will run as long as the `_running` attribute is
        # set to `True`. To safely quit the command, just set this attribute to `False` and the
        # command will finish a running tick and quit afterwards.
        self._running = True

        self.tick = 1  # every second

        # Register system signal handler to gracefully quit the service when
        # getting a `SIGINT` or `SIGTERM` signal (e.g. by CTRL+C).
        signal.signal(signal.SIGINT, self._handle_termination)
        signal.signal(signal.SIGTERM, self._handle_termination)



from django.apps import apps
from django.conf import settings
from django.test import SimpleTestCase

from drf_anonymous_login.models import AnonymousLogin
from testapp.models import PrivateModel, PublicModel


class TestSetup(SimpleTestCase):
    def test_installed_apps(self):
        self.assertIn("drf_anonymous_login", settings.INSTALLED_APPS)

    def test_models(self):
        self.assertIs(
            apps.get_model("drf_anonymous_login", "AnonymousLogin"),
            AnonymousLogin,
        )
        self.assertIs(apps.get_model("testapp", "PublicModel"), PublicModel)
        self.assertIs(apps.get_model("testapp", "PrivateModel"), PrivateModel)



from datetime import timedelta

from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_403_FORBIDDEN

from drf_anonymous_login.authentication import AUTH_HEADER, AUTH_KEYWORD
from drf_anonymous_login.management.commands.cleanup_tokens import Command
from drf_anonymous_login.models import AnonymousLogin
from testapp.models import PrivateModel, PublicModel, User


class TestApi(TestCase):
    def setUp(self):
        super().setUp()
        self.anonymous_login = AnonymousLogin.objects.create(request_data={})
        PublicModel.objects.create(name="public model")
        PrivateModel.objects.create(name="private model")

    def login_header(self):
        return {AUTH_HEADER: f"{AUTH_KEYWORD} {self.anonymous_login.token}"}

    def test_no_default_expiration_datetime(self):
        self.assertIsNone(self.anonymous_login.expiration_datetime)

    def test_default_expiration_datetime(self):
        with self.settings(ANONYMOUS_LOGIN_EXPIRATION=15):
            today = timezone.now()
            self.anonymous_login = AnonymousLogin.objects.create(request_data={})
            self.assertIsNotNone(self.anonymous_login.expiration_datetime)
            self.assertGreater(self.anonymous_login.expiration_datetime, today)
            self.assertLess(
                self.anonymous_login.expiration_datetime,
                today + timedelta(minutes=16),
            )

    def test_no_login(self):
        """
        Assert that token is required only for protected viewsets (with AnonymousLoginAuthenticationModelViewSet),
        not for public ones
        :return:
        """
        # make sure public model can be accessed by anyone
        url = reverse("publicmodel-list")
        response = self.client.get(url)
        self.assertEqual(HTTP_200_OK, response.status_code, response.content)

        # make sure private model can not be accessed by anyone
        url = reverse("privatemodel-list")
        response = self.client.get(url)
        self.assertEqual(HTTP_403_FORBIDDEN, response.status_code, response.content)

    def test_anonymous_login_token(self):
        """
        Assert that token allows access
        :return:
        """
        url = reverse("privatemodel-list")
        response = self.client.get(url, **self.login_header())
        self.assertEqual(HTTP_200_OK, response.status_code, response.content)

    def test_anonymous_login_token_invalid(self):
        """
        Assert that an invalid token does not allow access
        :return:
        """
        url = reverse("privatemodel-list")
        response = self.client.get(
            url,
            **{AUTH_HEADER: "{} {}".format(AUTH_KEYWORD, "invalid_token")},
        )
        self.assertEqual(HTTP_403_FORBIDDEN, response.status_code, response.content)
        self.assertEqual(str(response.data["detail"]), "Invalid authentication token")

    def test_anonymous_login_token_expired(self):
        """
        Assert that an expired token does not allow access any more
        :return:
        """
        self.anonymous_login.expiration_datetime = timezone.now() - timedelta(days=2)
        self.anonymous_login.save()
        url = reverse("privatemodel-list")
        response = self.client.get(url, **self.login_header())
        self.assertEqual(HTTP_403_FORBIDDEN, response.status_code, response.content)
        self.assertEqual(str(response.data["detail"]), "Expired authentication token")

    def test_anonymous_login_token_cleanup(self):
        """
        Assert that only an expired token gets deleted from the clean command
        :return:
        """
        self.assertEqual(AnonymousLogin.objects.count(), 1)

        cleanup_tokens = Command()

        # make sure cleanup does nothing
        cleanup_tokens.handle_tick()
        self.assertEqual(AnonymousLogin.objects.count(), 1)

        # expire the token
        self.anonymous_login.expiration_datetime = timezone.now() - timedelta(days=2)
        self.anonymous_login.save()

        # make sure the token gets deleted
        cleanup_tokens.handle_tick()
        self.assertEqual(AnonymousLogin.objects.count(), 0)

    def test_user_is_anonymous_login(self):
        """
        Assert that User is correctly identified as AnonymousLogin
        :return:
        """
        user = User.objects.create(
            username=self.anonymous_login.token,
            password="password",
        )
        self.assertTrue(user.is_anonymous_login)

    def test_user_is_not_anonymous_login(self):
        """
        Assert that User is correctly identified as no AnonymousLogin
        :return:
        """
        user = User.objects.create(username="user", password="password")
        self.assertFalse(user.is_anonymous_login)

    def test_user_get_anonymous_login(self):
        """
        Assert that User can access their AnonymousLogin
        :return:
        """
        user = User.objects.create(
            username=self.anonymous_login.token,
            password="password",
        )
        self.assertEqual(user.anonymous_login, self.anonymous_login)

    def test_user_get_no_anonymous_login(self):
        """
        Assert that User can not access an AnonymousLogin
        :return:
        """
        user = User.objects.create(username="user", password="password")
        self.assertIsNone(user.anonymous_login)

    def test_anonymous_login_with_cookie_authentication(self):
        """
        Assert that cookie authentication works
        :return:
        """
        # create anonymous login and check if cookie is set
        url = reverse("auth_anonymous-list")
        response = self.client.post(url)
        self.assertEqual(HTTP_201_CREATED, response.status_code)
        self.assertEqual(
            response.cookies["anonymous_token"].value,
            f"Token {response.data['token']}",
        )

        # set cookies and make sure it works
        url = reverse("privatemodel-list")
        self.client.cookies.load({"anonymous_token": f"Token {response.data['token']}"})
        response = self.client.get(url)
        self.assertEqual(HTTP_200_OK, response.status_code)



# Generated by Django 3.2.16 on 2022-12-12 15:07

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AnonymousLogin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "token",
                    models.CharField(
                        db_index=True,
                        max_length=64,
                        unique=True,
                        verbose_name="Token",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created"),
                ),
                ("request_data", models.JSONField(default=dict)),
            ],
        ),
    ]



from rest_framework import serializers

from testapp.models import PrivateModel, PublicModel


class PublicModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = PublicModel
        fields = [
            "name",
        ]


class PrivateModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = PrivateModel
        fields = [
            "name",
        ]



import binascii
import os
from datetime import timedelta

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class AnonymousLogin(models.Model):
    token = models.CharField(_("token"), max_length=64, db_index=True, unique=True)
    created = models.DateTimeField(_("created"), auto_now_add=True)
    request_data = models.JSONField(default=dict)
    expiration_datetime = models.DateTimeField(
        _("expiration datetime"),
        null=True,
        default=None,
    )

    def save(self, *args, **kwargs):
        if not self.token:
            self.token = self.generate_token()

        if not self.expiration_datetime:
            self.expiration_datetime = self.set_default_expiration_datetime()
        return super().save(*args, **kwargs)

    @staticmethod
    def generate_token():
        return binascii.hexlify(os.urandom(32)).decode()

    @staticmethod
    def set_default_expiration_datetime():
        default_expiration = getattr(settings, "ANONYMOUS_LOGIN_EXPIRATION", None)
        if default_expiration:
            return timezone.now() + timedelta(minutes=default_expiration)


class AnonymousLoginUserMixin:
    @property
    def is_anonymous_login(self):
        return AnonymousLogin.objects.filter(token=self.username).exists()

    @property
    def anonymous_login(self):
        """
        Returns the "longest" (with the latest expiration) AnonymousLogin element matching the user's username
        """
        return AnonymousLogin.objects.filter(token=self.username).order_by("-expiration_datetime").first()
